# coding=utf-8
# Copyright 2020 Rafael Jesús Castaño Ribes. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A NAF Agent.
Implements the Continuous Q-Learning Algorithm with Normalized Advantage
Functions (NAF), from "Continuous Deep Q-Learning with Model-based 
Acceleration" - Gu, S. et al. 2016
"""

import tensorflow as tf

from tf_agents.agents.tf_agent import TFAgent
from tf_agents.typing import types
from tf_agents.networks import network
from tf_agents.trajectories import time_step

from tf_agents.policies import actor_policy

import tf_agents

import numpy as np

from tf_agents.environments.tf_environment import TFEnvironment

from tf_agents.agents.tf_agent import LossInfo

from tf_agents.networks.value_network import ValueNetwork
from tf_agents.networks import network

from tf_agents.policies import gaussian_policy
from tf_agents.policies import ou_noise_policy
from tf_agents.policies import epsilon_greedy_policy

from tf_agents.replay_buffers.tf_uniform_replay_buffer import TFUniformReplayBuffer

from tf_agents.trajectories import trajectory
from tf_agents.trajectories import time_step
from tf_agents.trajectories import policy_step 

from tf_agents.utils import common
from tf_agents.utils import nest_utils

from typing import Optional, Text

import tensorflow_probability as tfp

class NafAgent(TFAgent):
    """A NAF Agent."""

    def __init__(self,
                 time_step_spec: time_step.TimeStep, 
                 action_spec: types.NestedTensorSpec, 
                 v_network: network.Network, 
                 l_matrix_network: network.Network, 
                 policy_network: network.Network, 
                 optimizer: types.Optimizer,
                 target_v_network: Optional[network.Network] = None, 
                 target_update_tau: types.Float = 1.0, 
                 target_update_period: types.Int = 1, 
                 td_errors_loss_fn: types.LossFn = tf.math.squared_difference, 
                 gamma: types.Float = 1.0, 
                 train_step_counter: Optional[tf.Variable] = None, 
                 noise_factor: types.Float = 0.1,
                 debug_summaries: types.Bool = False, 
                 summarize_grads_and_vars: types.Bool = False, 
                 name: Optional[Text] = None
            ):

        """Creates a NAF Agent.
        Args:
            time_step_spec: A `TimeStep` spec of the expected time_steps.
            action_spec: A nest of BoundedTensorSpec representing the actions.
            v_network: A tf_agents.network.Network to be used by the agent to 
            predict the v-value of an observation. It will be called with 
            call(observation, step_type, network_state) and should return 
            the estimated v_value.
            l_matrix_network: A tf_agents.network.Network to be used by the 
            agent to obtain the terms of the L matrix used to compute the 
            advantage value on an action. It will be called with 
            call(observation, step_type, network_state) and should return 
            the elements to build the matrix.
            policy_network: A tf_agents.network.Network to be used by the 
            agent to obtain the action to execute given an observation of 
            the environment. It will be called with call(observation, 
            step_type [,policy_state]) and should return the action to execute.
            optimizer: The optimizer to use for fit the networks.
            target_v_network: (Optional.)  A `tf_agents.network.Network` to 
            be used during Q learning to predict the target V values. 
            Every `target_update_period` train steps, the weights from 
            `v_network` are copied (possibly withsmoothing via 
            `target_update_tau`) to `target_v_network`. If `target_v_network` 
            is not provided, it is created by making a copy of the `v_network`, 
            which initializes a new network with the same structure and its 
            own layers and weights. 
            Performing a `Network.copy` does not work when the network instance
            already has trainable parameters (e.g., has already been built, or
            when the network is sharing layers with another).  In these cases, it is
            up to you to build a copy having weights that are not
            shared with the original `actor_network`, so that this can be used as a
            target network.  If you provide a `target_actor_network` that shares any
            weights with `actor_network`, a warning will be logged but no exception
            is thrown.
            target_update_tau: Factor for soft update of the target networks.
            target_update_period: Period for soft update of the target networks.
            td_errors_loss_fn:  A function for computing the TD errors loss. If None,
            a default value of elementwise huber_loss is used.
            gamma: A discount factor for future rewards.
            train_step_counter: An optional counter to increment every time the train
            op is run.  Defaults to the global_step.
            noise_factor: Standard deviation for the Gaussian noise added
            in the default collect policy.
            debug_summaries: A bool to gather debug summaries.
            summarize_grads_and_vars: If True, gradient and network variable summaries
            will be written during training.
            name: The name of this agent. All variables in this module will fall
            under that name.
        """

        tf.Module.__init__(self, name=name)

        self._check_action_spec(action_spec)

        # Create the policy and the collection policy

        self._policy_network = policy_network

        policy = actor_policy.ActorPolicy(time_step_spec = time_step_spec, 
                                          action_spec = action_spec, 
                                          actor_network = self._policy_network, 
                                          clip=True)
        
        collect_policy = gaussian_policy.GaussianPolicy(policy, scale = noise_factor)
        # collect_policy = ou_noise_policy.OUNoisePolicy(policy, ou_stddev = noise_factor, ou_damping = 1.0)
        # collect_policy = epsilon_greedy_policy.EpsilonGreedyPolicy(policy, epsilon = noise_factor)
        # collect_policy = ou_noise_policy.OUNoisePolicy(epsilon_greedy_policy.EpsilonGreedyPolicy(policy, epsilon = noise_factor), ou_stddev = noise_factor, ou_damping = 1.0)

        super(NafAgent, self).__init__(
            time_step_spec=time_step_spec
            , action_spec=action_spec
            , policy=policy
            , collect_policy=collect_policy
            , train_sequence_length=2
            , train_step_counter=train_step_counter
            , debug_summaries=debug_summaries
            , summarize_grads_and_vars=summarize_grads_and_vars
            )

        # Randomly initialize normalized Q network Q(x,u|0_Q).

        self._v_network = v_network
        self._v_network.create_variables()

        self._l_matrix_network = l_matrix_network
        self._l_matrix_network.create_variables()

        self._policy_network = policy_network
        self._policy_network.create_variables()
        
        # Create the target network Q'
        
        # Target network is used to generate target values for leading learning
        # As only the target_v_network is used to compute target v values, we 
        # only need to maintain this target_v_network

        _target_v_network = target_v_network

        if _target_v_network:
            _target_v_network.create_variables()
        else:
            _target_v_network = self._v_network.copy()
            if self._v_network._shared_network:
                _target_preprocessing_network = self._v_network._shared_network.copy()
                _target_preprocessing_network.create_variables()
                _target_v_network.set_shared_network(_target_preprocessing_network)
                _target_v_network.create_variables()

        self._target_v_network = common.maybe_copy_target_network_with_checks(
            self._v_network
            , _target_v_network
            , 'TargetVNetwork')

        # Initialize target network Q' with weight 0_Q' <- 0_Q

        common.soft_variables_update(
            self._v_network.variables, 
            self._target_v_network.variables, 
            tau=1.0)

        # Store other needed parameters

        self._target_update_tau = target_update_tau
        self._target_update_period = target_update_period
        self._update_target = self._get_target_updater(
            tau=self._target_update_tau
            , period=self._target_update_period)
        
        self._gamma = gamma

        self._optimizer = optimizer
        self._td_errors_loss_fn = td_errors_loss_fn
        

    def _initialize(self):

        # Initialize target network Q' with weight 0_Q' <- 0_Q
        common.soft_variables_update(
            self._v_network.variables, 
            self._target_v_network.variables, 
            tau=1.0)

    def _check_action_spec(self, action_spec):
        """Attribution:
            This method has been adapted from the _train method of the SacAgent class from the TF-Agents lib:
            https://github.com/tensorflow/agents/blob/v0.6.0/tf_agents/agents/sac/sac_agent.py
            Copyright 2020 The TF-Agents Authors. Licensed under the Apache License, Version 2.0.
        """
        flat_action_spec = tf.nest.flatten(action_spec)
        for spec in flat_action_spec:
          if spec.dtype.is_integer:
            raise NotImplementedError(
                'NafAgent does not currently support discrete actions. '
                'Action spec: {}'.format(action_spec))

    def _train(self, experience, weights=None):
        """This method trains the agent with the provided batched experience.
        Args:
          experience: A batch of trajectories.
          weights: Optional scalar or elementwise (per-batch-entry) importance
            weights.
        Returns:
          A LossInfo object, with the average loss obtained
        Attribution:
            This method has been adapted from the _train method of the SacAgent class from the TF-Agents lib:
            https://github.com/tensorflow/agents/blob/v0.6.0/tf_agents/agents/sac/sac_agent.py
            Copyright 2020 The TF-Agents Authors. Licensed under the Apache License, Version 2.0.
        """

        # Get the batches in the form of tensors for each concept
        time_steps, action_steps, next_time_steps = (trajectory.experience_to_transitions(experience, squeeze_time_dim=True))
        
        actions = action_steps.action
        
        # Get the variables to train, deduplicating the shared ones
        variables_to_train = list(
            common.object_identity.ObjectIdentitySet(
                self._v_network.trainable_variables +
                self._l_matrix_network.trainable_variables +
                self._policy_network.trainable_variables))

        # Record the gradient while calculating the loss
        with tf.GradientTape(persistent=True) as tape:
            # Ensure the gradient tape is watching the variables to train
            tape.watch(variables_to_train)
            # Compute the loss L
            loss_info = self._loss(
                time_steps, 
                actions, 
                next_time_steps, 
                td_errors_loss_fn=self._td_errors_loss_fn,
                tape=tape,
                gamma=self._gamma,
                training=True)
        
        # Get the gradients
        grads = tape.gradient(loss_info.loss, variables_to_train)
        
        # Pack gradients and variables together
        grads_and_vars = list(zip(grads, variables_to_train)) # list(...) is required for Python3.
        
        if self._summarize_grads_and_vars:
            # Summarize this info if requested, in order to analyze it in TensorBoard
            eager_utils.add_variables_summaries(grads_and_vars, self.train_step_counter)
            eager_utils.add_gradients_summaries(grads_and_vars, self.train_step_counter)

        # Apply the gradients with the optimizer function to optimize the agent networks
        self._optimizer.apply_gradients(grads_and_vars)
        
        # Increase the train step counter
        self.train_step_counter.assign_add(1)

        # Update the target network: 0_Q' <- tau*0_Q + (1 - tau)*0_Q'
        self._update_target()

        return loss_info

    def _loss(self,
                first_time_step: time_step.TimeStep, 
                action: policy_step.ActionType,
                next_time_step: time_step.TimeStep,
                td_errors_loss_fn: types.LossFn,
                tape: tf.GradientTape,
                gamma: types.Float = 1.0,
                training: bool = False) -> types.Tensor:

        with tf.name_scope('loss'):

            with tape.stop_recording():

                valid_transition_mask = tf.logical_and(
                    first_time_step.step_type != time_step.StepType.LAST,
                    next_time_step.step_type != time_step.StepType.FIRST
                    )

                terminal_valid_transition_mask = tf.logical_and(
                    next_time_step.step_type == time_step.StepType.LAST,
                    valid_transition_mask
                    )
                
                non_terminal_valid_transition_mask = tf.logical_and(
                    next_time_step.step_type == time_step.StepType.MID,
                    valid_transition_mask
                    )

                non_terminal_transition_observation = tf.boolean_mask(first_time_step.observation, non_terminal_valid_transition_mask)
                non_terminal_transition_action = tf.boolean_mask(action, non_terminal_valid_transition_mask)
                non_terminal_transition_next_observation = tf.boolean_mask(next_time_step.observation, non_terminal_valid_transition_mask)
                non_terminal_transition_reward = tf.boolean_mask(next_time_step.reward, non_terminal_valid_transition_mask)
                
                terminal_transition_observation = tf.boolean_mask(first_time_step.observation, terminal_valid_transition_mask)
                terminal_transition_action = tf.boolean_mask(action, terminal_valid_transition_mask)
                terminal_transition_reward = tf.boolean_mask(next_time_step.reward, terminal_valid_transition_mask)


                # Set y_i = r_i + gamma*V'(x'_i|0_Q')

                v_values = self._compute_target_v_values(non_terminal_transition_next_observation)
                discounted_v_values = tf.scalar_mul(gamma, v_values)

                non_terminal_y_i = tf.add(non_terminal_transition_reward, discounted_v_values)
                terminal_y_i = terminal_transition_reward

                y_i = tf.concat([non_terminal_y_i, terminal_y_i], 0)

                # Prepare data for Q computation

                valid_observations = tf.concat([non_terminal_transition_observation, terminal_transition_observation], 0)
                valid_actions = tf.concat([non_terminal_transition_action, terminal_transition_action], 0)

            # compute Q(x_i,u_i|0_Q)

            q_i = self._compute_current_q_values(valid_observations, valid_actions)

            # calculate loss: L = (1/N)*Sum_i[(y_i - Q(x_i,u_i|0_Q))^2]
            td_loss = td_errors_loss_fn(y_i, q_i)
            avg_loss = tf.reduce_mean(td_loss)

            losses_dict = {'avg_loss': avg_loss}

            common.summarize_scalar_dict(losses_dict, step=self.train_step_counter, name_scope='Losses/')

            if self._summarize_grads_and_vars:
                with tf.name_scope('Variables/VNetwork'):
                  for var in self._v_network.trainable_weights:
                    tf.compat.v2.summary.histogram(
                        name=var.name.replace(':', '_'),
                        data=var,
                        step=self.train_step_counter)
                with tf.name_scope('Variables/LNetwork'):
                  for var in self._l_matrix_network.trainable_weights:
                    tf.compat.v2.summary.histogram(
                        name=var.name.replace(':', '_'),
                        data=var,
                        step=self.train_step_counter)
                with tf.name_scope('Variables/PolicyNetwork'):
                  for var in self._policy_network.trainable_weights:
                    tf.compat.v2.summary.histogram(
                        name=var.name.replace(':', '_'),
                        data=var,
                        step=self.train_step_counter)

            if self._debug_summaries:
                
                diff_q_values = y_i - q_i
                td_error = tf.abs(diff_q_values)
                
                common.generate_tensor_summaries('target_q_values', y_i, self.train_step_counter)
                common.generate_tensor_summaries('current_q_values', q_i, self.train_step_counter)
                common.generate_tensor_summaries('diff_q_values', diff_q_values, self.train_step_counter)
                common.generate_tensor_summaries('td_error', td_error, self.train_step_counter)
                common.generate_tensor_summaries('td_loss', td_loss, self.train_step_counter)

        return LossInfo((avg_loss),())

    

    def _get_target_updater(self, tau=1.0, period=1):
        """Performs a soft update of the target network parameters.
        For each weight w_s in the v network, and its corresponding
        weight w_t in the target_v_network, a soft update is:
        w_t = (1 - tau) * w_t + tau * w_s
        Args:
            tau: A float scalar in [0, 1]. Default `tau=1.0` means hard update.
            period: Step interval at which the target network is updated.
        Returns:
            A callable that performs a soft update of the target network parameters.
        Attribution:
            This method has been adapted from:
            https://github.com/tensorflow/agents/blob/master/tf_agents/agents/dqn/dqn_agent.py#L359
            Copyright 2020 The TF-Agents Authors. Licensed under the Apache License, Version 2.0.
        """
        
        with tf.name_scope('update_targets'):

            def update():
                
                # As only the v_network is used to compute the target values, only this
                # network needs to be updated
                
                return common.soft_variables_update(
                    self._v_network.variables,
                    self._target_v_network.variables,
                    tau=tau,
                    tau_non_trainable=1.0)
            
            return common.Periodically(update, period, 'update_targets')


    def _compute_target_v_values(self, observation, training=False):
        v_values = self._target_v_network.call(observation, training)[0]
        return v_values

    def _compute_current_q_values(self, observation, action, training=False):
        # Q(x,u|0_Q) = A(x,u|0_A) + V(x|0_V)
        v_values = self._compute_current_v_values(observation, training)
        a_values = self._compute_current_a_values(observation, action, training)
        q_values = tf.add(v_values, a_values)
        return q_values

    def _compute_current_v_values(self, observation, training=False):
        v_values = self._v_network.call(observation, training)[0]
        return v_values
    
    def _compute_current_a_values(self, observation, action, training=False):
        # A(x,u|0_A) = ((-1/2)(u - policy(x|0_policy))^T)*P(x|0_P)*(u - policy(x|0_policy))
        time_step = tf_agents.trajectories.time_step.TimeStep(step_type=None, reward=None, discount=None, observation=observation)
        delta_action_transposed = tf.expand_dims(action, axis=1) - tf.expand_dims(self._policy.action(time_step).action, axis=1)
        delta_action = tf.linalg.matrix_transpose(delta_action_transposed)
        p_matrix = self._get_p_matrix(observation)
        aux = tf.linalg.matmul(delta_action_transposed, p_matrix)
        aux = tf.linalg.matmul(aux, delta_action)
        a_value = tf.squeeze(tf.math.scalar_mul(-0.5, aux), axis=[-1, -2])
        return a_value

    def _get_p_matrix(self, state):
        # P(x|0_P) = L(x|0_P)*L(x|0_p)^T
        l_matrix = self._get_l_matrix(state)
        return tf.linalg.matmul(l_matrix,tf.linalg.matrix_transpose(l_matrix))

    def _get_l_matrix(self, state):
        # L(x|0_P) is a lower-triangular matrix whose entries come from a linear output layer of a neural network, 
        # with the diagonal terms exponentiated
        l_matrix_terms = self._l_matrix_network.call(state, training=True)[0]
        pre_l_matrix = tfp.math.fill_triangular(l_matrix_terms)
        return tf.linalg.set_diag(pre_l_matrix, tf.math.exp(tf.linalg.diag_part(pre_l_matrix)))