import tensorflow as tf
from tf_agents.networks import network
from tf_agents.networks import utils as networks_utils
from tf_agents.networks import actor_distribution_network

class CompositeNetwork(network.Network):
    def __init__(self,
        input_tensor_spec,
        output_layer_params=1,
        fc_layer_params=(75, 40),
        dropout_layer_params=None,
        activation_fn=tf.keras.activations.relu,
        kernel_initializer=None,
        batch_squash=True,
        dtype=tf.float32,
        name='CompositeNetwork'):
        
        super(CompositeNetwork, self).__init__(
            input_tensor_spec=input_tensor_spec,
            state_spec=(),
            name=name)

        self._shared_network = None

        self._hidden_layers = networks_utils.mlp_layers(
            fc_layer_params=fc_layer_params,
            dropout_layer_params=dropout_layer_params,
            activation_fn=activation_fn,
            kernel_initializer=kernel_initializer,
            )

        self._output_layer = tf.keras.layers.Dense(
            output_layer_params,
            activation=None,
            kernel_initializer=tf.compat.v1.initializers.random_uniform(
                minval=-0.03, maxval=0.03))

    def set_shared_network(self, shared_network):
        if not self.built:
            if shared_network.built:
                self._shared_network = shared_network
            else:
                raise ValueError("The shared network must have been built")
        else:
            raise ValueError("You cannot initialize the network after being built")

    def call(self, observation, step_type=None, network_state=(), training=False):
        logits = observation
        current_network_state = network_state

        if self._shared_network is not None:
            logits, current_network_state = self._shared_network(logits, step_type=step_type, network_state=current_network_state, training=training)
        
        for layer in self._hidden_layers:
            logits = layer(logits, training=training)

        logits = self._output_layer(logits, training=training)

        return logits, current_network_state

    def create_variables(self, input_tensor_spec=None, **kwargs):
        next_layer_input_spec = None
        if self._shared_network is None:
            next_layer_input_spec = input_tensor_spec
        super(CompositeNetwork, self).create_variables(input_tensor_spec=next_layer_input_spec, **kwargs)

class CompositeValueNetwork(CompositeNetwork):
    def __init__(self,
        input_tensor_spec,
        fc_layer_params=(75, 40),
        dropout_layer_params=None,
        activation_fn=tf.keras.activations.relu,
        kernel_initializer=None,
        batch_squash=True,
        dtype=tf.float32,
        name='CompositeValueNetwork'
        ):
        super(CompositeValueNetwork, self).__init__(
            input_tensor_spec,
            output_layer_params=1,
            fc_layer_params=fc_layer_params,
            dropout_layer_params=dropout_layer_params,
            activation_fn=activation_fn,
            kernel_initializer=kernel_initializer,
            batch_squash=batch_squash,
            dtype=dtype,
            name=name
            )

    def call(self, observation, step_type=None, network_state=(), training=False):
        logits, current_network_state = super(CompositeValueNetwork, self).call(observation, step_type=step_type, network_state=network_state, training=training)
        logits = tf.squeeze(logits, -1)
        return logits, current_network_state

class CompositeLNetwork(CompositeNetwork):
    def __init__(self,
        input_tensor_spec,
        action_spec,
        fc_layer_params=(75, 40),
        dropout_layer_params=None,
        activation_fn=tf.keras.activations.relu,
        kernel_initializer=None,
        batch_squash=True,
        dtype=tf.float32,
        name='CompositeValueNetwork'
        ):

        # action_spec_size = Number of terms of the action space
        action_spec_size = tf.reduce_prod(action_spec.shape).numpy()

        # the number of terms o the L-Matrix is defined by the sum of the succession 
        # from 1 to the number of terms of the action space, this is:
        l_network_num_terms = int(action_spec_size * (action_spec_size + 1) / 2)

        super(CompositeLNetwork, self).__init__(
            input_tensor_spec,
            output_layer_params=l_network_num_terms,
            fc_layer_params=fc_layer_params,
            dropout_layer_params=dropout_layer_params,
            activation_fn=activation_fn,
            kernel_initializer=kernel_initializer,
            batch_squash=batch_squash,
            dtype=dtype,
            name=name
            )

class CompositeActorDistributionNetwork(network.Network):
    def __init__(self,
        input_tensor_spec,
        output_tensor_spec,
        fc_layer_params=(200, 100),
        dropout_layer_params=None,
        activation_fn=tf.keras.activations.relu,
        kernel_initializer=None,
        batch_squash=True,
        dtype=tf.float32,
        discrete_projection_net=actor_distribution_network._categorical_projection_net,
        continuous_projection_net=actor_distribution_network._normal_projection_net,
        name='CompositeActorDistributionNetwork'):
        
        super(CompositeActorDistributionNetwork, self).__init__(
            input_tensor_spec=input_tensor_spec,
            state_spec=(),
            name=name)

        self._shared_network = None

        
        self._ad_network_params = {
            "output_tensor_spec": output_tensor_spec,
            "fc_layer_params": fc_layer_params,
            "dropout_layer_params": dropout_layer_params,
            "activation_fn": activation_fn,
            "kernel_initializer": kernel_initializer,
            "batch_squash": batch_squash,
            "dtype": dtype,
            "discrete_projection_net": discrete_projection_net,
            "continuous_projection_net": continuous_projection_net
            }
        

        self._ad_network = None

    def initialize(self, shared_network):
        if not self.built:
            if shared_network.built:
                self._shared_network = shared_network
            else:
                raise ValueError("The shared network must have been built")
        else:
            raise ValueError("You cannot initialize the network after being built")

    def call(self, observation, step_type=None, network_state=(), training=False):
        logits = observation
        current_network_state = network_state

        if self._shared_network is not None:
            logits, current_network_state = self._shared_network(logits, step_type=step_type, network_state=current_network_state, training=training)
        
        logits, current_network_state = self._ad_network(logits, step_type=step_type, network_state=current_network_state, training=training)

        return logits, current_network_state

    def create_variables(self, input_tensor_spec=None, **kwargs):
        
        self_input_spec = None
        
        if self._shared_network is None:
            
            self_input_spec = input_tensor_spec

            if input_tensor_spec is None:
                ad_network_input_spec = self.input_tensor_spec
            else:
                ad_network_input_spec = input_tensor_spec
        
        else:
            ad_network_input_spec = self._shared_network._network_output_spec
        
        self._ad_network = actor_distribution_network.ActorDistributionNetwork(
            ad_network_input_spec,
            self._ad_network_params["output_tensor_spec"],
            fc_layer_params=self._ad_network_params["fc_layer_params"],
            dropout_layer_params=self._ad_network_params["dropout_layer_params"],
            activation_fn=self._ad_network_params["activation_fn"],
            kernel_initializer=self._ad_network_params["kernel_initializer"],
            batch_squash=self._ad_network_params["batch_squash"],
            dtype=self._ad_network_params["dtype"],
            discrete_projection_net=self._ad_network_params["discrete_projection_net"],
            continuous_projection_net=self._ad_network_params["continuous_projection_net"],
            name="ActorDistributionNetwork"
            )
        
        self._ad_network.create_variables(input_tensor_spec=ad_network_input_spec, **kwargs)

        super(CompositeActorDistributionNetwork, self).create_variables(input_tensor_spec=self_input_spec, **kwargs)