import numpy as np
import tensorflow as tf
import tf_agents.typing.types as types
from tf_agents.environments.py_environment import PyEnvironment
from tf_agents.environments.wrappers import PyEnvironmentBaseWrapper
from tf_agents.trajectories import time_step as ts
import tf_agents.specs.array_spec as array_spec

import collections

class ActionRepeatHistoryWrapper(PyEnvironmentBaseWrapper):
    """Repeates actions over n-steps while batching observations and accumulating the received reward.
       Attribution:
         This class is a fussion of the classes ActionRepeat and HistoryWrapper from the TF-Agents library:
           ActionRepeat   -> https://github.com/tensorflow/agents/blob/v0.6.0/tf_agents/environments/wrappers.py#L190-L221
           HistoryWrapper -> https://github.com/tensorflow/agents/blob/v0.6.0/tf_agents/environments/wrappers.py#L797-L877
         Both classes are copyrighted 2020 by the TF-Agents Authors and licensed under the Apache License, Version 2.0.
    """

    def __init__(self, env: PyEnvironment, times: types.Int):
        """Creates an action repeat wrapper.
        Args:
	        env: Environment to wrap.
	        times: Number of times the action should be executed. All the consequent observations will be batched and the reward will be the sum of the rewards.
        Raises:
	        ValueError: If the times parameter is not greater than 1.
        """
        super(ActionRepeatHistoryWrapper, self).__init__(env)

        if times < 0:
            raise ValueError('Times parameter ({}) should be greater than 0'.format(times))
        self._times = times
        self._history_length = times
        self._zero_observation = self._zeros_from_spec(env.observation_spec())
        self._observation_history = collections.deque(maxlen=self._history_length)
        self._former_observation_spec = env.observation_spec()
        self._observation_spec = self._get_observation_spec()

    def _get_observation_spec(self):
        return array_spec.add_outer_dims_nest(self._former_observation_spec, (self._times,))

    def observation_spec(self) -> types.NestedArraySpec:
        return self._observation_spec

    def _zeros_from_spec(self, spec):
        
        def _zeros(spec):
            return np.zeros(spec.shape, dtype=spec.dtype)

        return tf.nest.map_structure(_zeros, spec)

    def _add_history(self, time_step):
        self._observation_history.append(time_step.observation)

    def _reset(self):
        self._observation_history.extend([self._zero_observation] * (self._history_length - 1))
        
        time_step = self._env.reset()
        self._add_history(time_step)

        observation = tf.convert_to_tensor(self._observation_history)

        return ts.TimeStep(time_step.step_type, time_step.reward, time_step.discount, observation)

    def _step(self, action):

        if self.current_time_step() is None or self.current_time_step().is_last():
            return self._reset()

        total_reward = 0

        self._observation_history.clear()

        for step_number in range(self._times):

            time_step = self._env.step(action)
            
            if time_step.is_first():
                self._observation_history.extend([self._zero_observation] * (self._history_length - 1))
                self._add_history(time_step)
                break

            self._add_history(time_step)
            
            total_reward += time_step.reward

            if time_step.is_last():
                self._observation_history.extend([self._zero_observation] * (self._history_length - 1 - step_number))
                break
                

        total_reward = np.asarray(total_reward, dtype=np.asarray(time_step.reward).dtype)

        observation = tf.convert_to_tensor(self._observation_history)

        return ts.TimeStep(time_step.step_type, total_reward, time_step.discount, observation)
