# coding=utf-8
# Copyright 2020 The TF-Agents Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python2, python3
r"""Train and Eval SAC.
All hyperparameters come from the SAC paper
https://arxiv.org/pdf/1812.05905.pdf
To run:
```bash
tensorboard --logdir $HOME/tmp/sac/gym/HalfCheetah-v2/ --port 2223 &
python tf_agents/agents/sac/examples/v2/train_eval.py \
  --root_dir=$HOME/tmp/sac/gym/HalfCheetah-v2/ \
  --alsologtostderr
```
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import tensorflow as tf  # pylint: disable=g-explicit-tensorflow-version-import

from experiments.off_p_tsb_experiment import OffPolicyTimeStepBasedExperiment
from agents.naf.utils import CompositeValueNetwork
from agents.naf.utils import CompositeLNetwork
from agents.naf.utils import CompositeActorDistributionNetwork
from agents.naf.naf_agent import NafAgent
from tf_agents.agents.sac import tanh_normal_projection_network
from tf_agents.networks import encoding_network
from tf_agents.agents.tf_agent import TFAgent

class NafExperiment(OffPolicyTimeStepBasedExperiment):
    """A simple train and eval class for a NAF Agent."""
    def __init__(self,
                 root_dir = "",
                 num_eval_episodes = 10,
                 summary_interval = 1000,
                 env_name = "BipedalWalker-v3",
                 env_action_repeat_times = 4,
                 preprocessing_conv_layer_params = None,
                 preprocessing_conv_type = encoding_network.CONV_TYPE_1D,
                 preprocessing_fc_layer_params = (256, ),
                 preprocessing_dropout_layer_params = None,
                 v_network_fc_layer_params = (256, ),
                 v_network_dropout_layer_params = None,
                 l_network_fc_layer_params = (256, ),
                 l_network_dropout_layer_params = None,
                 policy_network_fc_layer_params = (256, ),
                 policy_network_dropout_layer_params = None,
                 policy_network_uses_shared_preprocessing_network = True,
                 learning_rate=3e-4,
                 target_update_tau=0.005,
                 target_update_period=1,
                 td_errors_loss_fn=tf.math.squared_difference,
                 gamma=0.99,
                 noise_factor=0.1,
                 debug_summaries=False,
                 summarize_grads_and_vars=False,
                 replay_buffer_capacity=1000000,
                 initial_collect_steps=10000,
                 collect_steps_per_iteration=1,
                 use_tf_functions=True,
                 sample_batch_size = 256,
                 num_iterations=3000000,
                 train_steps_per_iteration=1,
                 log_interval=1000,
                 eval_interval=10000,
                 train_checkpoint_interval=50000,
                 policy_checkpoint_interval=50000,
                 replay_buffer_checkpoint_interval=50000,
                 name="default_naf_experiment"
                 ):

        super(NafExperiment, self).__init__(
            root_dir = root_dir,
            num_eval_episodes = num_eval_episodes,
            summary_interval = summary_interval,
            env_name = env_name,
            env_action_repeat_times = env_action_repeat_times,
            debug_summaries = debug_summaries,
            summarize_grads_and_vars = summarize_grads_and_vars,
            replay_buffer_capacity = replay_buffer_capacity,
            initial_collect_steps = initial_collect_steps,
            collect_steps_per_iteration = collect_steps_per_iteration,
            use_tf_functions = use_tf_functions,
            sample_batch_size = sample_batch_size,
            num_iterations = num_iterations,
            train_steps_per_iteration = train_steps_per_iteration,
            log_interval = log_interval,
            eval_interval = eval_interval,
            train_checkpoint_interval = train_checkpoint_interval,
            policy_checkpoint_interval = policy_checkpoint_interval,
            replay_buffer_checkpoint_interval = replay_buffer_checkpoint_interval,
            name = name
            )
        
        self.preprocessing_conv_layer_params = preprocessing_conv_layer_params
        self.preprocessing_conv_type = preprocessing_conv_type
        self.preprocessing_fc_layer_params = preprocessing_fc_layer_params
        self.preprocessing_dropout_layer_params = preprocessing_dropout_layer_params
        self.v_network_fc_layer_params = v_network_fc_layer_params
        self.v_network_dropout_layer_params = v_network_dropout_layer_params
        self.l_network_fc_layer_params = l_network_fc_layer_params
        self.l_network_dropout_layer_params = l_network_dropout_layer_params
        self.policy_network_fc_layer_params = policy_network_fc_layer_params
        self.policy_network_dropout_layer_params = policy_network_dropout_layer_params
        self.policy_network_uses_shared_preprocessing_network = policy_network_uses_shared_preprocessing_network,
        self.learning_rate = learning_rate
        self.target_update_tau = target_update_tau
        self.target_update_period = target_update_period
        self.td_errors_loss_fn = td_errors_loss_fn
        self.gamma = gamma
        self.noise_factor = noise_factor

    def _copy(self):
        return NafExperiment(
            root_dir = self.root_dir,
            num_eval_episodes = self.num_eval_episodes,
            summary_interval = self.summary_interval,
            env_name = self.env_name,
            env_action_repeat_times = self.env_action_repeat_times,
            preprocessing_conv_layer_params = self.preprocessing_conv_layer_params,
            preprocessing_conv_type = self.preprocessing_conv_type,
            preprocessing_fc_layer_params = self.preprocessing_fc_layer_params,
            preprocessing_dropout_layer_params = self.preprocessing_dropout_layer_params,
            v_network_fc_layer_params = self.v_network_fc_layer_params,
            v_network_dropout_layer_params = self.v_network_dropout_layer_params,
            l_network_fc_layer_params = self.l_network_fc_layer_params,
            l_network_dropout_layer_params = self.l_network_dropout_layer_params,
            policy_network_fc_layer_params = self.policy_network_fc_layer_params,
            policy_network_dropout_layer_params = self.policy_network_dropout_layer_params,
            policy_network_uses_shared_preprocessing_network = self.policy_network_uses_shared_preprocessing_network,
            learning_rate = self.learning_rate,
            target_update_tau = self.target_update_tau,
            target_update_period = self.target_update_period,
            td_errors_loss_fn = self.td_errors_loss_fn,
            gamma = self.gamma,
            noise_factor = self.noise_factor,
            debug_summaries = self.debug_summaries,
            summarize_grads_and_vars = self.summarize_grads_and_vars,
            replay_buffer_capacity = self.replay_buffer_capacity,
            initial_collect_steps = self.initial_collect_steps,
            collect_steps_per_iteration = self.collect_steps_per_iteration,
            use_tf_functions = self.use_tf_functions,
            sample_batch_size = self.sample_batch_size,
            num_iterations = self.num_iterations,
            train_steps_per_iteration = self.train_steps_per_iteration,
            log_interval = self.log_interval,
            eval_interval = self.eval_interval,
            train_checkpoint_interval = self.train_checkpoint_interval,
            policy_checkpoint_interval = self.policy_checkpoint_interval,
            replay_buffer_checkpoint_interval = self.replay_buffer_checkpoint_interval,
            name = self.name
            )

    def _get_the_agent(self, 
                       time_step_spec, 
                       observation_spec, 
                       action_spec, 
                       global_step, 
                       debug_summaries, 
                       summarize_grads_and_vars
                       ) -> TFAgent:

        # Q Function Networks

        # Shared Preprocessing Network (Optional)

        preprocessing_network = encoding_network.EncodingNetwork(
            observation_spec,
            conv_layer_params = self.preprocessing_conv_layer_params,
            conv_type = self.preprocessing_conv_type,
            fc_layer_params = self.preprocessing_fc_layer_params,
            dropout_layer_params = self.preprocessing_dropout_layer_params
            )

        preprocessing_network.create_variables()

        # V-value Network
        v_value_network = CompositeValueNetwork(
            observation_spec,
            fc_layer_params = self.v_network_fc_layer_params,
            dropout_layer_params = self.v_network_dropout_layer_params
            )
            
        v_value_network.set_shared_network(preprocessing_network)

        # L-Matrix Network
        l_matrix_network = CompositeLNetwork(
            observation_spec,
            action_spec,
            fc_layer_params  =self.l_network_fc_layer_params,
            dropout_layer_params = self.l_network_dropout_layer_params
            )

        l_matrix_network.set_shared_network(preprocessing_network)

        policy_network = CompositeActorDistributionNetwork(
            observation_spec,
            action_spec,
            fc_layer_params = self.policy_network_fc_layer_params,
            dropout_layer_params = self.policy_network_dropout_layer_params,
            continuous_projection_net = tanh_normal_projection_network.TanhNormalProjectionNetwork
            )

        if self.policy_network_uses_shared_preprocessing_network:
            policy_network.initialize(preprocessing_network)
            
        tf_agent = NafAgent(
            time_step_spec,
            action_spec,
            v_network=v_value_network,
            l_matrix_network=l_matrix_network,
            policy_network=policy_network,
            optimizer=tf.compat.v1.train.AdamOptimizer(
                learning_rate=self.learning_rate),
            target_update_tau=self.target_update_tau,
            target_update_period=self.target_update_period,
            td_errors_loss_fn=self.td_errors_loss_fn,
            gamma=self.gamma,
            noise_factor=self.noise_factor,
            debug_summaries=self.debug_summaries,
            summarize_grads_and_vars=self.summarize_grads_and_vars,
            train_step_counter=global_step)
            
        return tf_agent