# coding=utf-8
# Copyright 2020 The TF-Agents Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python2, python3
r"""Train and Eval SAC.
All hyperparameters come from the SAC paper
https://arxiv.org/pdf/1812.05905.pdf
To run:
```bash
tensorboard --logdir $HOME/tmp/sac/gym/HalfCheetah-v2/ --port 2223 &
python tf_agents/agents/sac/examples/v2/train_eval.py \
  --root_dir=$HOME/tmp/sac/gym/HalfCheetah-v2/ \
  --alsologtostderr
```
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import time

from absl import logging

from six.moves import range
import tensorflow as tf  # pylint: disable=g-explicit-tensorflow-version-import

from tf_agents.agents.tf_agent import TFAgent
from tf_agents.agents.random import random_agent

from experiments.off_p_tsb_experiment import OffPolicyTimeStepBasedExperiment

class RndExperiment(OffPolicyTimeStepBasedExperiment):
    """A simple train and eval class for a Random Policy Agent."""
    def __init__(self,
                 root_dir = "",
                 num_eval_episodes = 10,
                 summary_interval = 1000,
                 env_name = "BipedalWalker-v3",
                 env_action_repeat_times = 4,
                 debug_summaries=False,
                 summarize_grads_and_vars=False,
                 replay_buffer_capacity=1000000,
                 initial_collect_steps=10000,
                 collect_steps_per_iteration=1,
                 use_tf_functions=True,
                 sample_batch_size = 256,
                 num_iterations=3000000,
                 train_steps_per_iteration=1,
                 log_interval=1000,
                 eval_interval=10000,
                 train_checkpoint_interval=50000,
                 policy_checkpoint_interval=50000,
                 replay_buffer_checkpoint_interval=50000,
                 name="default_naf_experiment"
                 ):
        
        super(RndExperiment, self).__init__(
            root_dir = root_dir,
            num_eval_episodes = num_eval_episodes,
            summary_interval = summary_interval,
            env_name = env_name,
            env_action_repeat_times = env_action_repeat_times,
            debug_summaries = debug_summaries,
            summarize_grads_and_vars = summarize_grads_and_vars,
            replay_buffer_capacity = replay_buffer_capacity,
            initial_collect_steps = initial_collect_steps,
            collect_steps_per_iteration = collect_steps_per_iteration,
            use_tf_functions = use_tf_functions,
            sample_batch_size = sample_batch_size,
            num_iterations = num_iterations,
            train_steps_per_iteration = train_steps_per_iteration,
            log_interval = log_interval,
            eval_interval = eval_interval,
            train_checkpoint_interval = train_checkpoint_interval,
            policy_checkpoint_interval = policy_checkpoint_interval,
            replay_buffer_checkpoint_interval = replay_buffer_checkpoint_interval,
            name = name
            )
    
    def _copy(self):
        return RndExperiment(
            root_dir = self.root_dir,
            num_eval_episodes = self.num_eval_episodes,
            summary_interval = self.summary_interval,
            env_name = self.env_name,
            env_action_repeat_times = self.env_action_repeat_times,
            debug_summaries = self.debug_summaries,
            summarize_grads_and_vars = self.summarize_grads_and_vars,
            replay_buffer_capacity = self.replay_buffer_capacity,
            initial_collect_steps = self.initial_collect_steps,
            collect_steps_per_iteration = self.collect_steps_per_iteration,
            use_tf_functions = self.use_tf_functions,
            sample_batch_size = self.sample_batch_size,
            num_iterations = self.num_iterations,
            train_steps_per_iteration = self.train_steps_per_iteration,
            log_interval = self.log_interval,
            eval_interval = self.eval_interval,
            train_checkpoint_interval = self.train_checkpoint_interval,
            policy_checkpoint_interval = self.policy_checkpoint_interval,
            replay_buffer_checkpoint_interval = self.replay_buffer_checkpoint_interval,
            name = self.name
            )

    def _get_the_agent(self, time_step_spec, observation_spec, action_spec, global_step, debug_summaries, summarize_grads_and_vars) -> TFAgent:
        return random_agent.RandomAgent(time_step_spec, 
                                        action_spec, 
                                        train_step_counter=global_step, 
                                        debug_summaries=debug_summaries, 
                                        summarize_grads_and_vars=summarize_grads_and_vars,
                                        num_outer_dims=2)