# coding=utf-8
# Copyright 2020 The TF-Agents Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python2, python3
r"""Train and Eval SAC.
All hyperparameters come from the SAC paper
https://arxiv.org/pdf/1812.05905.pdf
To run:
```bash
tensorboard --logdir $HOME/tmp/sac/gym/HalfCheetah-v2/ --port 2223 &
python tf_agents/agents/sac/examples/v2/train_eval.py \
  --root_dir=$HOME/tmp/sac/gym/HalfCheetah-v2/ \
  --alsologtostderr
```
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import tensorflow as tf  # pylint: disable=g-explicit-tensorflow-version-import

from experiments.off_p_tsb_experiment import OffPolicyTimeStepBasedExperiment

from tf_agents.agents.ddpg import critic_network
from tf_agents.agents.sac import sac_agent
from tf_agents.agents.sac import tanh_normal_projection_network
from tf_agents.networks import actor_distribution_network

from tf_agents.agents.tf_agent import TFAgent

class SacExperiment(OffPolicyTimeStepBasedExperiment):
    """A simple train and eval class for a SAC agent."""
    def __init__(self,
                 root_dir = "",
                 num_eval_episodes = 10,
                 summary_interval = 1000,
                 env_name = "BipedalWalker-v3",
                 env_action_repeat_times = 4,
                 actor_fc_layers=(256, 256),
                 critic_obs_fc_layers=None,
                 critic_action_fc_layers=None,
                 critic_joint_fc_layers=(256, 256),
                 actor_learning_rate=3e-4,
                 critic_learning_rate=3e-4,
                 alpha_learning_rate=3e-4,
                 target_update_tau=0.005,
                 target_update_period=1,
                 td_errors_loss_fn=tf.math.squared_difference,
                 gamma=0.99,
                 reward_scale_factor=0.1,
                 gradient_clipping=None,
                 debug_summaries=False,
                 summarize_grads_and_vars=False,
                 replay_buffer_capacity=1000000,
                 initial_collect_steps=10000,
                 collect_steps_per_iteration=1,
                 use_tf_functions=True,
                 sample_batch_size = 256,
                 num_iterations=3000000,
                 train_steps_per_iteration=1,
                 log_interval=1000,
                 eval_interval=10000,
                 train_checkpoint_interval=50000,
                 policy_checkpoint_interval=50000,
                 replay_buffer_checkpoint_interval=50000,
                 name="sac_default_experiment"
                 ):

        super(SacExperiment, self).__init__(
            root_dir = root_dir,
            num_eval_episodes = num_eval_episodes,
            summary_interval = summary_interval,
            env_name = env_name,
            env_action_repeat_times = env_action_repeat_times,
            debug_summaries = debug_summaries,
            summarize_grads_and_vars = summarize_grads_and_vars,
            replay_buffer_capacity = replay_buffer_capacity,
            initial_collect_steps = initial_collect_steps,
            collect_steps_per_iteration = collect_steps_per_iteration,
            use_tf_functions = use_tf_functions,
            sample_batch_size = sample_batch_size,
            num_iterations = num_iterations,
            train_steps_per_iteration = train_steps_per_iteration,
            log_interval = log_interval,
            eval_interval = eval_interval,
            train_checkpoint_interval = train_checkpoint_interval,
            policy_checkpoint_interval = policy_checkpoint_interval,
            replay_buffer_checkpoint_interval = replay_buffer_checkpoint_interval,
            name = name
            )
        
        self.actor_fc_layers = actor_fc_layers
        self.critic_obs_fc_layers = critic_obs_fc_layers
        self.critic_action_fc_layers = critic_action_fc_layers
        self.critic_joint_fc_layers = critic_joint_fc_layers
        self.actor_learning_rate = actor_learning_rate
        self.critic_learning_rate = critic_learning_rate
        self.alpha_learning_rate = alpha_learning_rate
        self.target_update_tau = target_update_tau
        self.target_update_period = target_update_period
        self.td_errors_loss_fn = td_errors_loss_fn
        self.gamma = gamma
        self.reward_scale_factor = reward_scale_factor
        self.gradient_clipping = gradient_clipping

    def _copy(self):
        return SacExperiment(
            root_dir = self.root_dir,
            num_eval_episodes = self.num_eval_episodes,
            summary_interval = self.summary_interval,
            env_name = self.env_name,
            env_action_repeat_times = self.env_action_repeat_times,
            actor_fc_layers = self.actor_fc_layers,
            critic_obs_fc_layers = self.critic_obs_fc_layers,
            critic_action_fc_layers = self.critic_action_fc_layers,
            critic_joint_fc_layers = self.critic_joint_fc_layers,
            actor_learning_rate = self.actor_learning_rate,
            critic_learning_rate = self.critic_learning_rate,
            alpha_learning_rate = self.alpha_learning_rate,
            target_update_tau = self.target_update_tau,
            target_update_period = self.target_update_period,
            td_errors_loss_fn = self.td_errors_loss_fn,
            gamma = self.gamma,
            reward_scale_factor = self.reward_scale_factor,
            gradient_clipping = self.gradient_clipping,
            debug_summaries = self.debug_summaries,
            summarize_grads_and_vars = self.summarize_grads_and_vars,
            replay_buffer_capacity = self.replay_buffer_capacity,
            initial_collect_steps = self.initial_collect_steps,
            collect_steps_per_iteration = self.collect_steps_per_iteration,
            use_tf_functions = self.use_tf_functions,
            sample_batch_size = self.sample_batch_size,
            num_iterations = self.num_iterations,
            train_steps_per_iteration = self.train_steps_per_iteration,
            log_interval = self.log_interval,
            eval_interval = self.eval_interval,
            train_checkpoint_interval = self.train_checkpoint_interval,
            policy_checkpoint_interval = self.policy_checkpoint_interval,
            replay_buffer_checkpoint_interval = self.replay_buffer_checkpoint_interval,
            name = self.name
            )

    def _get_the_agent(self, 
                       time_step_spec, 
                       observation_spec, 
                       action_spec, 
                       global_step, 
                       debug_summaries, 
                       summarize_grads_and_vars
                       ) -> TFAgent:

        actor_net = actor_distribution_network.ActorDistributionNetwork(
            observation_spec,
            action_spec,
            fc_layer_params=self.actor_fc_layers,
            continuous_projection_net=tanh_normal_projection_network.TanhNormalProjectionNetwork)
            
        critic_net = critic_network.CriticNetwork(
            (observation_spec, action_spec),
            observation_fc_layer_params=self.critic_obs_fc_layers,
            action_fc_layer_params=self.critic_action_fc_layers,
            joint_fc_layer_params=self.critic_joint_fc_layers,
            kernel_initializer='glorot_uniform',
            last_kernel_initializer='glorot_uniform')

        tf_agent = sac_agent.SacAgent(
            time_step_spec,
            action_spec,
            actor_network=actor_net,
            critic_network=critic_net,
            actor_optimizer=tf.compat.v1.train.AdamOptimizer(
                learning_rate=self.actor_learning_rate),
            critic_optimizer=tf.compat.v1.train.AdamOptimizer(
                learning_rate=self.critic_learning_rate),
            alpha_optimizer=tf.compat.v1.train.AdamOptimizer(
                learning_rate=self.alpha_learning_rate),
            target_update_tau=self.target_update_tau,
            target_update_period=self.target_update_period,
            td_errors_loss_fn=self.td_errors_loss_fn,
            gamma=self.gamma,
            reward_scale_factor=self.reward_scale_factor,
            gradient_clipping=self.gradient_clipping,
            debug_summaries=self.debug_summaries,
            summarize_grads_and_vars=self.summarize_grads_and_vars,
            train_step_counter=global_step)

        return tf_agent