# coding=utf-8
# Copyright 2020 Rafael Jesús Castaño Ribes.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Train and Eval class for off-policy TD learning agents feeded with batches of experiences from a replay buffer.
Attribution: 
  This class is an adaptation of the module train_eval.py from the SAC Agent examples from the TF-Agents library:
    https://github.com/tensorflow/agents/blob/v0.6.0/tf_agents/agents/sac/examples/v2/train_eval.py
  Copyrighted 2020 by the TF-Agents Authors and licensed under the Apache License, Version 2.0.
Changes added to the original code:
  The code has been given a class format and has been generalized to work with any off-policy TD based TF-Agent, wich will have to be provided by 
  subclasses via the method _get_the_agent().
  Moreover, a method called create_video has been added allowing the class to generate a video by playing the obtained exploitation policy on the environment. This method has been adapted from the example code provided by 
  Holt, S. (2020). Value and Policy based agents: https://pathtopioneer.com/blog/2020/07/rl-4, licensed under a Creative Commons CC BY-NC-SA 4.0 license.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import time

from absl import logging

from six.moves import range
import tensorflow as tf  # pylint: disable=g-explicit-tensorflow-version-import

from tf_agents.agents.tf_agent import TFAgent
from tf_agents.agents.random import random_agent
from tf_agents.drivers import dynamic_step_driver

from tf_agents.environments import suite_gym
from environments.wrappers import ActionRepeatHistoryWrapper
from tf_agents.environments import tf_py_environment

from tf_agents.eval import metric_utils
from tf_agents.metrics import tf_metrics

from tf_agents.policies import greedy_policy
from tf_agents.policies import random_tf_policy

from tf_agents.replay_buffers import tf_uniform_replay_buffer
from tf_agents.utils import common

import imageio

class OffPolicyTimeStepBasedExperiment(object):
    """A simple train and eval class for Off-Policy TimeStep based Agents"""
    def __init__(self,
                 root_dir = "",
                 num_eval_episodes = 10,
                 summary_interval = 1000,
                 env_name = "BipedalWalker-v3",
                 env_action_repeat_times = 4,
                 debug_summaries=False,
                 summarize_grads_and_vars=False,
                 replay_buffer_capacity=1000000,
                 initial_collect_steps=10000,
                 collect_steps_per_iteration=1,
                 use_tf_functions=True,
                 sample_batch_size = 256,
                 num_iterations=3000000,
                 train_steps_per_iteration=1,
                 log_interval=1000,
                 eval_interval=10000,
                 train_checkpoint_interval=50000,
                 policy_checkpoint_interval=50000,
                 replay_buffer_checkpoint_interval=50000,
                 name="default_naf_experiment"
                 ):
        
        self.root_dir = root_dir
        self.num_eval_episodes = num_eval_episodes
        self.summary_interval = summary_interval
        self.env_name = env_name
        self.env_action_repeat_times = env_action_repeat_times
        self.debug_summaries = debug_summaries
        self.summarize_grads_and_vars = summarize_grads_and_vars
        self.replay_buffer_capacity = replay_buffer_capacity
        self.initial_collect_steps = initial_collect_steps
        self.collect_steps_per_iteration = collect_steps_per_iteration
        self.use_tf_functions = use_tf_functions
        self.sample_batch_size = sample_batch_size
        self.num_iterations = num_iterations
        self.train_steps_per_iteration = train_steps_per_iteration
        self.log_interval = log_interval
        self.eval_interval = eval_interval
        self.train_checkpoint_interval = train_checkpoint_interval
        self.policy_checkpoint_interval = policy_checkpoint_interval
        self.replay_buffer_checkpoint_interval = replay_buffer_checkpoint_interval
        self.name = name

    def copy(self):
        return self._copy()
    
    def _copy(self):
        return OffPolicyTimeStepBasedExperiment(
            root_dir = self.root_dir,
            num_eval_episodes = self.num_eval_episodes,
            summary_interval = self.summary_interval,
            env_name = self.env_name,
            env_action_repeat_times = self.env_action_repeat_times,
            debug_summaries = self.debug_summaries,
            summarize_grads_and_vars = self.summarize_grads_and_vars,
            replay_buffer_capacity = self.replay_buffer_capacity,
            initial_collect_steps = self.initial_collect_steps,
            collect_steps_per_iteration = self.collect_steps_per_iteration,
            use_tf_functions = self.use_tf_functions,
            sample_batch_size = self.sample_batch_size,
            num_iterations = self.num_iterations,
            train_steps_per_iteration = self.train_steps_per_iteration,
            log_interval = self.log_interval,
            eval_interval = self.eval_interval,
            train_checkpoint_interval = self.train_checkpoint_interval,
            policy_checkpoint_interval = self.policy_checkpoint_interval,
            replay_buffer_checkpoint_interval = self.replay_buffer_checkpoint_interval,
            name = self.name
            )

    def launch(self):

        tf.compat.v1.enable_v2_behavior()
        logging.set_verbosity(logging.INFO)

        base_dir = os.path.join(self.root_dir, self.name)

        train_dir = os.path.join(base_dir, 'train')
        eval_dir = os.path.join(base_dir, 'eval')

        train_summary_writer = tf.compat.v2.summary.create_file_writer(train_dir)
        train_summary_writer.set_as_default()

        eval_summary_writer = tf.compat.v2.summary.create_file_writer(eval_dir)
        
        eval_metrics = [
            tf_metrics.AverageReturnMetric(buffer_size=self.num_eval_episodes),
            tf_metrics.AverageEpisodeLengthMetric(buffer_size=self.num_eval_episodes)
        ]

        global_step = tf.compat.v1.train.get_or_create_global_step()

        with tf.compat.v2.summary.record_if(lambda: tf.math.equal(global_step % self.summary_interval, 0)):
            
            py_env = ActionRepeatHistoryWrapper(suite_gym.load(self.env_name), times=self.env_action_repeat_times)
            tf_env = tf_py_environment.TFPyEnvironment(py_env)

            eval_py_env = ActionRepeatHistoryWrapper(suite_gym.load(self.env_name), times=self.env_action_repeat_times)
            eval_tf_env = tf_py_environment.TFPyEnvironment(py_env)

            time_step_spec = tf_env.time_step_spec()
            observation_spec = time_step_spec.observation
            action_spec = tf_env.action_spec()

            # Get the agent; Subclasses must override the _get_the_agent method
            tf_agent = self._get_the_agent(time_step_spec,
                                           observation_spec, 
                                           action_spec, 
                                           global_step,
                                           self.debug_summaries,
                                           self.summarize_grads_and_vars) 

            tf_agent.initialize()

            # Make the replay buffer.
            replay_buffer = tf_uniform_replay_buffer.TFUniformReplayBuffer(
                data_spec=tf_agent.collect_data_spec,
                batch_size=tf_env.batch_size,
                max_length=self.replay_buffer_capacity)

            replay_observer = [replay_buffer.add_batch]

            train_metrics = [
                tf_metrics.NumberOfEpisodes(),
                tf_metrics.EnvironmentSteps(),
                tf_metrics.AverageReturnMetric(
                    buffer_size=self.num_eval_episodes, batch_size=tf_env.batch_size),
                tf_metrics.AverageEpisodeLengthMetric(
                    buffer_size=self.num_eval_episodes, batch_size=tf_env.batch_size),
            ]

            eval_policy = greedy_policy.GreedyPolicy(tf_agent.policy) if not isinstance(tf_agent, random_agent.RandomAgent) else tf_agent.policy
            
            initial_collect_policy = random_tf_policy.RandomTFPolicy(time_step_spec, action_spec)
            
            collect_policy = tf_agent.collect_policy

            train_checkpointer = common.Checkpointer(
                ckpt_dir=train_dir,
                agent=tf_agent,
                global_step=global_step,
                metrics=metric_utils.MetricsGroup(train_metrics, 'train_metrics'))

            policy_checkpointer = common.Checkpointer(
                ckpt_dir=os.path.join(train_dir, 'policy'),
                policy=eval_policy,
                global_step=global_step)

            replay_buffer_checkpointer = common.Checkpointer(
                ckpt_dir=os.path.join(train_dir, 'replay_buffer'),
                max_to_keep=1,
                replay_buffer=replay_buffer)

            train_checkpointer.initialize_or_restore()
            replay_buffer_checkpointer.initialize_or_restore()

            initial_collect_driver = dynamic_step_driver.DynamicStepDriver(
                tf_env,
                initial_collect_policy,
                observers=replay_observer + train_metrics,
                num_steps=self.initial_collect_steps)

            collect_driver = dynamic_step_driver.DynamicStepDriver(
                tf_env,
                collect_policy,
                observers=replay_observer + train_metrics,
                num_steps=self.collect_steps_per_iteration)

            if self.use_tf_functions:
                initial_collect_driver.run = common.function(initial_collect_driver.run)
                collect_driver.run = common.function(collect_driver.run)
                tf_agent.train = common.function(tf_agent.train)

            if replay_buffer.num_frames() == 0:
                # Collect initial replay data.
                logging.info(
                    'Initializing replay buffer by collecting experience for %d steps '
                    'with a random policy.', self.initial_collect_steps)
                initial_collect_driver.run()

            results = metric_utils.eager_compute(
                eval_metrics,
                eval_tf_env,
                eval_policy,
                num_episodes=self.num_eval_episodes,
                train_step=global_step,
                summary_writer=eval_summary_writer,
                summary_prefix='Metrics',
            )
            
            metric_utils.log_metrics(eval_metrics)

            time_step = None
            
            policy_state = collect_policy.get_initial_state(tf_env.batch_size)

            timed_at_step = global_step.numpy()
            
            time_acc = 0

            # Prepare replay buffer as dataset with invalid transitions filtered.
            def _filter_invalid_transition(trajectories, unused_arg1):
                return ~trajectories.is_boundary()[0]
                
            dataset = replay_buffer.as_dataset(
                sample_batch_size = self.sample_batch_size, 
                num_steps = 2
                ).unbatch().filter(
                    _filter_invalid_transition
                    ).batch(
                        self.sample_batch_size
                        ).prefetch(5)
                
            # Dataset generates trajectories with shape [Bx2x...]
            iterator = iter(dataset)

            def train_step():
                experience, _ = next(iterator)
                return tf_agent.train(experience)

            if self.use_tf_functions:
                train_step = common.function(train_step)

            global_step_val = global_step.numpy()

            while global_step_val < self.num_iterations:

                start_time = time.time()
                time_step, policy_state = collect_driver.run(
                    time_step=time_step,
                    policy_state=policy_state,
                )
                
                for _ in range(self.train_steps_per_iteration):
                    train_loss = train_step()
                    time_acc += time.time() - start_time

                global_step_val = global_step.numpy()

                if global_step_val % self.log_interval == 0:
                    
                    logging.info('step = %d, loss = %f', global_step_val, train_loss.loss)
                
                    steps_per_sec = (global_step_val - timed_at_step) / time_acc
                
                    logging.info('%.3f steps/sec', steps_per_sec)

                    tf.compat.v2.summary.scalar(name='global_steps_per_sec', data=steps_per_sec, step=global_step)

                    timed_at_step = global_step_val
                    
                    time_acc = 0

                for train_metric in train_metrics:
                    train_metric.tf_summaries(train_step=global_step, step_metrics=train_metrics[:2])

                if global_step_val % self.eval_interval == 0:
                    results = metric_utils.eager_compute(
                        eval_metrics,
                        eval_tf_env,
                        eval_policy,
                        num_episodes=self.num_eval_episodes,
                        train_step=global_step,
                        summary_writer=eval_summary_writer,
                        summary_prefix='Metrics',
                        )

                if global_step_val % self.log_interval == 0:
                    metric_utils.log_metrics(eval_metrics)

                if global_step_val % self.train_checkpoint_interval == 0:
                    train_checkpointer.save(global_step=global_step_val)

                if global_step_val % self.policy_checkpoint_interval == 0:
                    policy_checkpointer.save(global_step=global_step_val)

                if global_step_val % self.replay_buffer_checkpoint_interval == 0:
                    replay_buffer_checkpointer.save(global_step=global_step_val)

        video_file_name = os.path.join(self.root_dir, self.name + ".mp4")
        self._create_video(video_filename=video_file_name, py_env=py_env, agent=tf_agent)
        
        return train_loss

    def _create_video(self, video_filename="video.mp4", num_episodes = 3, py_env=None, agent=None):
        """Create a video for a few episodes
           Attribution:
             This method has been adapted from the example code provided by 
               Holt, S. (2020). Value and Policy based agents: https://pathtopioneer.com/blog/2020/07/rl-4
               licensed under a Creative Commons CC BY-NC-SA 4.0 license.
        """
        env = tf_py_environment.TFPyEnvironment(py_env)
        with imageio.get_writer(video_filename, fps=15) as video:
            for _ in range(num_episodes):
                time_step = env.reset()
                video.append_data(py_env.render())
                while not time_step.is_last():
                    action_step = agent.policy.action(time_step)
                    time_step = env.step(action_step.action)
                    video.append_data(py_env.render())

    def _get_the_agent(self, time_step_spec, observation_spec, action_spec, global_step, debug_summaries, summarize_grads_and_vars) -> TFAgent:
        # Subclasess must override this method"
        raise ValueError("This is a base class, you must implement a subclass and override this method")